"""OpenOT2 — Modular OT-2 Robot Control with ML/CV Integration."""

from openot2.client import OT2Client
from openot2.utils import setup_logging

__all__ = ["OT2Client", "setup_logging"]
__version__ = "0.1.0"
